/*
 *	Functions for analysing unknown systems
 *	ie. generic functions.
 */

#include "../../globals.h"
#include "../functions_allos.h"
#include "../allos.h"
#include "magic.h"
#include "../../arch/functions_allarch.h"

/*
 * Hook that perform automagic guess
 * of operating system, if first search fail.
 * performed only if user did not specified 
 * operating system or process name
 *
 * unknown_hook_proc_determine (pointer to process address)
 * return 0 on success 
 */

int unknown_hook_proc_determine(u_int64_t *process){
	
	dbgprint("process: %llX",*process);
	
	if ( *process != 0) {
		// We found MAGIC_PROCESS. 
		// If user did not specified magic_process or OS, 
		// then we now it is linux.
		if (magic_process_set == 0){	
			// Set OS as linux
			set_os_from_list("linux");
		}
	
		return 0;
	}


	if (magic_process_set != 0 ) {
		printf("	proc_determine: Cannot find process,"\
			 "exiting.\n"); 
		return 2;
	}

	// if user did not used --magic-process option
	// Try all again with windows process
	// This takes twice long, but we have "automatic" detection of OS;)

	fprintf(stderr,"	proc_determine: \"%s\" not found, "\
		"second try: \"%s\"\n", 
		img1.magic_process, 
		WINDOWS_MAGIC_PROCESS);

	search_exit(); //close file

	// Use windows criteria for process matching
	set_os_from_list("windows");

	// This seems to work, even if it is not optimal
	if (search_init(inputfile,WINDOWS_MAGIC_PROCESS,0) != 0 ) { 
		printf("proc_determine: search_init returned error. "\
			"Aborting.\n");
		return -2;
	}

	while (( *process = search_next()) != 0){
		dbgprint("Trying to verify: %llX as task_struct\n", *process);
		if (verify_task_struct(*process) == 0) { 
			// we have found valid process
			printf("Found valid windows process at 0x%llX.\n",
				*process);
			break;
		}
	}
	if (*process == 0) { //Found nothing (again)
		printf("	proc_determine: Cannot find process, "\
			"exiting.\n");
		// not windows => set OS back to unknown
		set_os_from_list("unknown");
		return 2;
	}

	return 0;
}

/*
 * Hook to determine aditional information,
 * once process structure is found.
 *
 * unknown_hook_proc_determine2 ()
 * return 0 on success 
 */

int unknown_hook_proc_determine2(){

	int status;

	// Find some userspace process
	if ( (status=lin_find_userspace_process()) !=0) {
		// Desperate heuristic, if process is stil unknown.
		// TODO remove? XXX BSD
		// no init process => probably win? 
		// this does not have to be true, 
		// but don't waste our results and treat it as win.
		printf("	linux_determine: lin_find_userspace_process "\
			"returned: %d, assuming windows OS.\n", status);
		set_os_from_list("windows");
		return 0;
	}

	// determine distances of VMAs
	if ( (status=lin_find_process_VMA(img1.ts_one_of_lists)) !=0) {
		printf("	linux_determine: lin_find_process_VMA "\
			"returned: %d, terminating.\n",status);
		printf("	linux_determine: Reading of userspace "\
			"addresses will be not available..\n");
		return 4;
	}

	// if VMA was sucessfully found, it is clear now that OS is linux
	set_os_from_list("linux");


	// VMAs are determined before process VPT,
	// because finding mm_struct offset inside task_struct is more reliable


	// determine VPT distances
	if ( (status=lin_find_process_VPT(img1.ts_one_of_lists)) !=0) {
		printf("	linux_determine: lin_find_process_VPT "\
			"returned: %d.\n",status);
		printf("	linux_determine: Reading of userspace "\
			"addresses will be not available..\n");
		return 6;
	}

	return 0;
}

/*
 * Hook for listing processes
 *
 * At this point, os should be no more "unknown"
 * But lets assume it is.
 *
 * unknown_hook_list_processes (pointer to VMA switch)
 * return 0 on success 
 */

int unknown_hook_list_processes(int *list_areas){

	if (do_areas_list) {
		printf("Disabling VMA listing, "\
			"not implemented for unknown systems.\n");
		*list_areas=0;
	}

	return 0;
}

/*
 * Hook for matching process structure
 *
 * linux_hook_verify_task_struct ( 
	pointer to hl size, 
	index in hl list,
	ver_addr (needed for bsd list,
	pointer to bsd list start
	pointer to found_os)
 * return 0 on success 
 */

int unknown_hook_verify_task_struct(int *hl_size, int i, u_int64_t ver_addr, u_int64_t *bsd_list_start, int *found_os) {

	int cyc_hl_size=0;
	int bsd_hl_size=0; 
	int bsd_list_strings_size=0;
	int max_bsd_list_strings_size=0;
	u_int64_t bsd_list_start_tmp=0;

	// Non-Cyclic (BSD) style list_head 
	bsd_hl_size=bsd_list_head_size(img1.ts_headlists[i],
			&bsd_list_start_tmp,
			ver_addr - img1.ts_headlists[i] + img1.pointer_size,
			&bsd_list_strings_size);
	dbgprint("BSD list string size: %d",bsd_list_strings_size);

	// Cyclic (Lin/Win) style list_head 
	cyc_hl_size=lin_list_head_size(img1.ts_headlists[i]);

	if (max(cyc_hl_size,bsd_hl_size) >= *hl_size) {
		// more and not loop (loop < 0)
		/* Sometimes, there are multiple pointers to list_head
		   we are looking for.
		   Therefore our used offset can be wrong, 
		   and we should count string for each element of list.
		 */

		dbgprint("Saving local maximum:(1.try) 0x%llX",
				img1.ts_headlists[i]);

		if (	(max(cyc_hl_size, bsd_hl_size) > *hl_size) || 
			(bsd_list_strings_size > max_bsd_list_strings_size)) { 

			dbgprint("Saving local maximum: 0x%llX, circ_hl: %d, \
					bsd_hl: %d", img1.ts_headlists[i], 
					cyc_hl_size, bsd_hl_size);

			//save which list_head is right one
			img1.ts_headlists_number=i; 

			if (cyc_hl_size >= bsd_hl_size) {
				// remember local maximum
				*hl_size=cyc_hl_size;		
				// list_head type
				*found_os=OS_CIRC_HL;	
			} else {
				// remember local maximum
				*hl_size=bsd_hl_size;		
				// list_head type
				*found_os=OS_BSD;	
				//remember first entry of list
				*bsd_list_start=bsd_list_start_tmp; 
				max_bsd_list_strings_size=bsd_list_strings_size;
			}
		}
	}

	return 0;
}

/*
 * Hook for matching process structure
 *
 * unknown_hook_verify_task_struct2 ( 
 *	found_os, start of non-cyclic hl, pointer to hl )
 *
 * return 0 on success 
 */

int unknown_hook_verify_task_struct2(int found_os, u_int64_t bsd_list_start, u_int64_t *list_start) {

	// We found list of some type, so set OS if unknown
	img1.os=found_os;

	// special case, bsd list is not circular
	// so we need to point to the first entry
	if ((img1.os==OS_BSD) && (bsd_list_start != 0)) {
		*list_start = bsd_list_start;
	}
	return 0;

}

/*
 * Function that sets pointers to all other 
 * os dependent functions and hooks
 *
 * unknown_init (pointer to os structure)
 * return 0 on success 
 */

int unknown_init(struct os_struct *self){

	strncpy(self->operating_system,"unknown\0",OS_NAME_LEN);
	strncpy(self->magic_process_name,UNKNOWN_MAGIC_PROCESS,TS_COMM_LEN);
	strncpy(self->magic_user_process_name,UNKNOWN_MAGIC_USER_PROCESS,TS_COMM_LEN);
	strncpy(self->magic_module_name,UNKNOWN_MAGIC_MODULE,MD_NAME_LEN);

	self->hook_proc_determine = (int (*)())unknown_hook_proc_determine;
	self->hook_proc_determine2 = (int (*)())unknown_hook_proc_determine2;
	self->hook_list_processes = (int (*)())unknown_hook_list_processes;
	self->hook_verify_task_struct = (int (*)())unknown_hook_verify_task_struct;
	self->hook_verify_task_struct2 = (int (*)())unknown_hook_verify_task_struct2;

	self->ts_magic_headlist_number = UNKNOWN_TS_MAGIC_HEADLIST_NUMBER;
	self->ts_magic_headlist_spread = UNKNOWN_TS_MAGIC_HEADLIST_SPREAD;
	self->ts_magic_headlist_front_number = UNKNOWN_TS_MAGIC_HEADLIST_FRONT_NUMBER;
	self->ts_magic_min_list_size = UNKNOWN_TS_MAGIC_MIN_LIST_SIZE;
	self->ts_magic_max_tries = UNKNOWN_TS_MAGIC_MAX_TRIES;
	self->ts_magic_hl_mm_dst_p = UNKNOWN_TS_MAGIC_HL_MM_DST_P;
	self->ts_magic_list_max_err = UNKNOWN_TS_MAGIC_LIST_MAX_ERR;
	self->ts_magic_headlist_max = UNKNOWN_TS_MAGIC_HEADLIST_MAX;
	self->mm_magic_pgd_dst_p = UNKNOWN_MM_MAGIC_PGD_DST_P;
	self->md_magic_headlist_number = UNKNOWN_MD_MAGIC_HEADLIST_NUMBER;
	self->md_magic_max_tries = UNKNOWN_MD_MAGIC_MAX_TRIES;
	self->md_magic_hl_name_dst_p = UNKNOWN_MD_MAGIC_HL_NAME_DST_P;
	self->md_magic_headlist_max = UNKNOWN_MD_MAGIC_HEADLIST_MAX;
	self->md_pattern_spread = UNKNOWN_MD_PATTERN_SPREAD;
	self->max_list_size = UNKNOWN_MAX_LIST_SIZE;



	return 0;
}


